# speed of light (m s^-1)
c = 2.99792458e8

# little g (m s^-2)
g = 9.806651

# Planck's constant (kg m^2 s)
h = 6.62607015e-34

# Reduced Planck's constant (kg m^2 s)
hbar = h / 6.283185307179586

# Boltzmann's constant (kg m^2 s^-2 K^-1)
k_B = 1.380649e-23

# CMB temperature (K)
T_CMB = 2.72548

MARIA_MIN_NU = 1e6  # 1 MHz
MARIA_MAX_NU = 15e12  # 150 THz
